unit unit_tttext_math;

interface

function TermToReal(s:string): extended;

implementation

uses unit15rechenblatt, sysutils;

(* (c) Joachim Mohr, Rottenburg am Neckar
   Die Unit darf frei in nichtkommerziellen Programmen verwendet
   werden, wenn der CopyRight-Vermerk nicht entfernt wird.
   Kritik, Anregungen, Verbesserungsvorschlge bitte an
   http://delphi.zsg-rottenburg.de *)

Procedure Fehlerbehandlung(const s: string);
begin
   //form15rechenblatt.Everlauf.text := s;
end;

function ln0(x:extended):extended;
begin
   if x<1E-15 then Begin
     result := 0;
     Fehlerbehandlung('Logarithmus');
   End else result := ln(x)
end;


function IsInteger(const x: extended; eps_Genauigkeit: extended): boolean;
begin
  result := frac(abs(x) + eps_Genauigkeit) < eps_Genauigkeit * 2;
end;

function HochInt(X: Extended; n: Integer): Extended;
begin
  if n < 0 then result := 1.0 / HochInt(x, -n) else Begin //Iterativ
    Result := 1.0;
    while n > 0 do BEgin
      while not Odd(n) do
      BEGin
        n := n shr 1;
        X := X * X
      END;
      Dec(n);
      Result := Result * X
    ENd;
  End;
end;

function HochReal(x, y: Extended): Extended;
begin
  result := 0;
  try
    if y = 0.0 then Result := 1.0 {x^0=1 0^x=0}
    else if (x = 0.0) and (y > 0) then Result := 0.0
    else if IsInteger(y, 1E-18) and (Abs(y) <= MaxInt) then
      Result := HochInt(x, Integer(Round(y)))
    else if x<=0 then Fehlerbehandlung('Potenzieren')
      else Result := Exp(y * Ln(x))
  except Fehlerbehandlung('Potenzieren') end;
end;

function tan(x: extended): extended;
begin
  try
    result := sin(x)/cos(x)
  except Fehlerbehandlung('tan'); result := 0 End;
end;

function arctan0(x: extended): extended;
begin
  try
    result := arctan(x)
  except Fehlerbehandlung('arctan'); result := 0 End;
end;

function ArcTan2(Y, X: Extended): Extended;
asm
        FLD     Y
        FLD     X
        FPATAN
        FWAIT
end;

function ArcCos0(X: Extended): Extended;
begin
  try
    Result := ArcTan2(Sqrt(1 - X * X), X);
  except result := 0; Fehlerbehandlung('arccos'); End;
end;

function ArcSin0(X: Extended): Extended;
begin
  try
    Result := ArcTan2(X, Sqrt(1 - X * X))
  except result := 0; Fehlerbehandlung('arcsin'); End;
end;


function division(x,y: extended): extended; //result := x/y
begin
  try
    result := x/y
  except Fehlerbehandlung('Division'); result := 0 End;
end;

function sqrt0(x: extended): extended;
begin
  try
    result := sqrt(x)
  except Fehlerbehandlung('sqrt'); result := 0 End;
end;

function int0(x:extended):extended; //= Gau'sche Klammerf.
begin
  result := int(x);
  if result > x then result := result - 1;
end;

function fakultaet(x: Extended): Extended;
begin
  if x <= 1 then result := 1 else  result := x*fakultaet(x - 1); //natrlich rekursiv
end;

function TermToRealMitX(s:string; x:extended): extended; //Darber: die mathematischen Hilfsfunktionen
   var xglobal: extended; //weitere Parameter denkbar
 function TTR(s:string): extended; //TTR = "Termtoreal ohne x"
  //  {Bisher '+', '-', '*', '/', '^', Klammern und 'x' integriert,
  //   d.h. gebrochen rationale Funktionen werden ausgewertet
     // Hilfsfunktionen 
      var u2,v2, u3,v3, u4, v4, u6, v6: string; //fr Funktionen wie "ln",  "sin", "sqrt"
   function pos0(c:char;s:string):integer;
     //pos0 findet das Zeichen "+","-" ... nicht innerhalb von Klammern
      var k,z:integer; //z:=Anzahl der Klammern
   begin
     z:=0;
     for k:=length(s) downto 1 do Begin //Korrigiert Dez. 2002
       if s[k]='(' then inc(z);
       if s[k]=')' then dec(z);
       if (z=0) and (s[k]=c) then BEgin
        result:=k; //Treffer
        exit;
      ENd;
    End;
    result:=0; //nichts gefunden
  end;

   function anfang(s:string;c:char):string;
   begin
     anfang:=copy(s,1,pos0(c,s)-1);
   end;

   function copyab(const s:string; const i:integer):string;
     begin result:=copy(s,i,length(s)-i+1) end;

   function ende(s:string; c:char):string;
   begin
     ende:=copyab(s,pos0(c,s)+1)
   end;
   Procedure MalzeichenSetzten(var s:string);
    //macht aus 2x = 2*x , aus 2(a+b) = 2*(a+b), aus (a+b)c = (a+b)*c,
    // aus (a+b)(a-b) =(a+b)*(a-b), aus 2sin(x) = 2*sin(x) u.s.e.
     var k: integer;
   begin
     for k := 1 to length(s) - 1 do
       if (s[k] in ['0'..'9',')']) and (s[k+1] in ['a'..'z','A'..'Z','(']) then Begin
       s := copy(s,1,k) + '*' + copyab(s,k+1);
       MalzeichenSetzten(s); //rekursiv
       exit; //length(s) ist grer geworden
    End;
end;
 begin
  result := 0;
  s := trim(s);
  if s = '' then exit;
  if s[1]='-' then s:='0'+s; //zB. s='-7/3x+14' -> s='0-7/3x+14'
  MalzeichenSetzten(s);
  u2:=copy(s,1,2);  //zum Beispiel u2 = 'ln'
  v2:=copyab(s,3);
  u3:=copy(s,1,3);  //zum Beispiel u3 = 'sin'
  v3:=copyab(s,4);
  u4:=copy(s,1,4);  //zum Beispiel u4 = 'sqrt'
  v4:=copyab(s,5);
  u6:=copy(s,1,6);  //zum Beispiel u4 = 'arctan'
  v6:=copyab(s,7);
  //Zuerst ganzrationale Funktion
  if pos0('+',s)>0  then result:=TTR(anfang(s,'+'))+TTR(ende(s,'+')) else
  if pos0('-',s)>0  then result:=TTR(anfang(s,'-'))-TTR(ende(s,'-')) else
  if pos0('*',s)>0 then  result:=TTR(anfang(s,'*'))*TTR(ende(s,'*')) else
  if pos0('',s)>0 then  result:=TTR(anfang(s,''))*TTR(ende(s,'')) else
  if pos0('/',s)>0 then  result:=division(TTR(anfang(s,'/')),TTR(ende(s,'/'))) else
  if pos0(':',s)>0 then  result:=division(TTR(anfang(s,':')),TTR(ende(s,':'))) else
  if pos0('^',s)>0 then  result:=hochreal(TTR(anfang(s,'^')),TTR(ende(s,'^'))) else
  //Jetzt die Funktionen
  if u2='ln' then result:=ln0(TTR(v2)) else
  if u2='lg' then result:=ln0(TTR(v2))/ln(10) else
  if u2='lb' then result:=ln0(TTR(v2))/ln(2) else
  if u3='sin' then result:=sin(TTR(v3)) else
  if u3='cos' then result:=cos(TTR(v3)) else
  if u3='tan' then result:=tan(TTR(v3)) else
  if u6='arctan' then result:=arctan0(TTR(v6)) else
  if u6='arcsin' then result:=arcsin0(TTR(v6)) else
  if u6='arccos' then result:=arccos0(TTR(v6)) else
  if u3='si_' then result:=sin(Pi/180*TTR(v3)) else
  if u3='co_' then result:=cos(Pi/180*TTR(v3)) else
  if u3='ta_' then result:=tan(Pi/180*TTR(v3)) else
  if u6='arcta_' then result:=arctan0(TTR(v6))*180/Pi else
  if u6='arcsi_' then result:=arcsin0(TTR(v6))*180/Pi else
  if u6='arcco_' then result:=arccos0(TTR(v6))*180/Pi else
  if u3='abs' then result:=abs(TTR(v3)) else
  if u3='exp' then result:=exp(TTR(v3)) else
  if u3='fak' then result:=fakultaet(TTR(v3)) else
  if u3='int' then result:=int0(TTR(v3)) else
  if u4='sqrt' then result:=sqrt0(TTR(v4)) else
  //Jetzt die Klammern
  if (s>'') and (s[1]='(') then Begin //Am Anfang und Ende eine Klammer
    s:=copy(s,2,length(s)-2);
    result:=TTR(s)
  End else
  if s='x' then result:=xGlobal else
    result:=StrToFloat(s);
 end;
begin //
  xGlobal := x;
  result := TTR(s);
end;

function KommaGleichPunktUndOhneLeerzeichen(const s: string): string;
  var k: integer;
begin
  result := '';
  for k := 1 to length(s) do if s[k] <> ' ' then Begin
    if s[k] in ['.',','] then result := result + Decimalseparator else
    result := result + s[k];
  End;
end;


function TermToReal(s:string): extended;
begin
  result := TermToRealmitX(KommaGleichPunktUndOhneLeerzeichen(s),0);
end;



end.
